#
# -*- coding: utf-8 -*-

"""
    unicode_ids
    ~~~~~~~~~~~

    The Sphinx extension to allow HTML builder to use non-ASCII ids.

    :copyright: (c) 2011-2013, 2015 by Suzumizaki-Kimitaka(about Additional works)
    :license: 2-clause BSD.
"""

"""The extension non-ascii id support for Sphinx

Both HTML4 and 5 accepts non-ASCII unicode characters used in url,
includes anchor.

But currently docutils (0.9) force the identifiers to ascii limited,
like #id-1234 or just snipping Unicode characters. Currently, the
docutils cannot treat non-ASCII characters.

Today, this behavior is just NONSENSE at least using HTML.
This module hacks docutils to enable
File_name_written_with_unicode#anchor_written_with_unicode.

This module hacks 'sphinx.environment.BuildEnvironment.resolve_toctree'.
And because of this, just cover characters included in the set of
sys.getdefaultencoding(), not all of Unicode characters.

For Windows, you might create 'usercustomize.py' or 'sitecustomize.py'
to call sys.setdefaultencoding('mbcs').
See PEP 370 http://www.python.org/dev/peps/pep-0370/ to know where 
to place the script.

This hack should be rewritten especially the time to go Python 3.

Hack work is done by Suzumizaki-Kimitaka, from May 2011.
Work for Sphinx 1.2b3 is done December 2013, we don't have to replace
'sphinx.environment.BuildEnvironment.resolve_toctree' for now. 

Updated at Mar 2015, for Python 3.4.3 and Sphinx 1.3
"""

#
# the imports below need to define alternate functions. copied from:
# environment.py
#

import os
import sys
from os import path
from docutils import nodes
from sphinx import addnodes
from sphinx.util \
    import url_re, get_matching_docs, docname_join, FilenameUniqDict

#
# to define new builder.
#
import sphinx.builders.html

#
# to support my_make_id() function
#
import re
import docutils

#
# to define alternate functions.
#
def to_unicode(s):
    """Get unicode object by decoding given string with filesystem encoding
    
    @param s the string
    @return unicode object, same as given 's' when unicode object given
    
    When the system encoding is not found, just return the value of 's'.
    """
    if sys.version_info[0] >= 3:
        return s
    if isinstance(s, str):
        # NOTE: I don't know how to perform with cygwin.
        if sys.platform == "windows":
            encoding = sys.getdefaultencoding()
        else:
            encoding = sys.getfilesystemencoding()
        if not encoding:
            return s
        return s.decode(encoding, "replace")
    return s

#
# alternate functions.
#

# ====================================================================
# alter 'docutils.nodes.make_id'
# ====================================================================

if len(u'\U0010FFFF') == 1:
    _unusable_characters_for_id = re.compile(u'[^0-9A-Za-z_\\-\u00a0-\U0010FFFF]')
else:
    _unusable_characters_for_id = re.compile(u'[^0-9A-Za-z_\\-\u00a0-\uFFFF]')

_unusable_prefixes = re.compile(u'^([0-9]|--)')

def my_make_id(s):
    """Return identifier keeping Unicode characters.

    Defined to replace the function 'docutils.nodes.make_id'.
    
    Currently, this function can't determine the return value will be used
    in HTML or CSS. Unusable characters will be replaced '_xhh' form where
    'hh' is 2-digit hexadecimal lowercase.
    """
    s = to_unicode(s.lower())
    s = u'-'.join(s.split()) # replace white spaces with hyphens.
    s = _unusable_characters_for_id.sub(lambda mo: u'_x{:02x}'.format(ord(mo.group())), s)
    if _unusable_prefixes.match(s):
        s = u'_' + s
    return s # we need unicode object, NOT str.

# ====================================================================
# alter 'sphinx.directives.other.TocTree.run'
# ====================================================================

from sphinx.util.nodes import explicit_title_re, set_source_info

def my_toctree_run(self):
    """Show non existing entries of toctree

    Defined to replace the function 'sphinx.directives.other.TocTree.run'

    Only %r following are replaced %s to avoid unreadable string.
    """
    env = self.state.document.settings.env
    suffixes = env.config.source_suffix
    glob = 'glob' in self.options
    caption = self.options.get('caption')
    if caption:
        self.options.setdefault('name', nodes.fully_normalize_name(caption))

    ret = []
    # (title, ref) pairs, where ref may be a document, or an external link,
    # and title may be None if the document's title is to be used
    entries = []
    includefiles = []
    all_docnames = env.found_docs.copy()
    # don't add the currently visited file in catch-all patterns
    all_docnames.remove(env.docname)
    for entry in self.content:
        if not entry:
            continue
        if glob and ('*' in entry or '?' in entry or '[' in entry):
            patname = docname_join(env.docname, entry)
            docnames = sorted(patfilter(all_docnames, patname))
            for docname in docnames:
                all_docnames.remove(docname)  # don't include it again
                entries.append((None, docname))
                includefiles.append(docname)
            if not docnames:
                ret.append(self.state.document.reporter.warning(
                    'toctree glob pattern %r didn\'t match any documents'
                    % entry, line=self.lineno))
        else:
            # look for explicit titles ("Some Title <document>")
            m = explicit_title_re.match(entry)
            if m:
                ref = m.group(2)
                title = m.group(1)
                docname = ref
            else:
                ref = docname = entry
                title = None
            # remove suffixes (backwards compatibility)
            for suffix in suffixes:
                if docname.endswith(suffix):
                    docname = docname[:-len(suffix)]
                    break
            # absolutize filenames
            docname = docname_join(env.docname, docname)
            if url_re.match(ref) or ref == 'self':
                entries.append((title, ref))
            elif docname not in env.found_docs:
                ret.append(self.state.document.reporter.warning(
                    'toctree contains reference to nonexisting '
                    'document %s' % docname, line=self.lineno))
                env.note_reread()
            else:
                all_docnames.discard(docname)
                entries.append((title, docname))
                includefiles.append(docname)
    subnode = addnodes.toctree()
    subnode['parent'] = env.docname
    # entries contains all entries (self references, external links etc.)
    subnode['entries'] = entries
    # includefiles only entries that are documents
    subnode['includefiles'] = includefiles
    subnode['maxdepth'] = self.options.get('maxdepth', -1)
    subnode['caption'] = caption
    subnode['glob'] = glob
    subnode['hidden'] = 'hidden' in self.options
    subnode['includehidden'] = 'includehidden' in self.options
    subnode['numbered'] = self.options.get('numbered', 0)
    subnode['titlesonly'] = 'titlesonly' in self.options
    set_source_info(self, subnode)
    wrappernode = nodes.compound(classes=['toctree-wrapper'])
    wrappernode.append(subnode)
    self.add_name(wrappernode)
    ret.append(wrappernode)
    return ret

def setup(app):
    """Extend Sphinx as we want

    @param app sphinx.application.Sphinx object to use add builder or so.
    """

    #
    # reset dirs to unicode, otherwise several 'replace(SEP, os.sep)'
    # codes make bad string with MBCS.  
    #
    app.srcdir = to_unicode(app.srcdir)
    app.confdir = to_unicode(app.confdir)
    app.outdir = to_unicode(app.outdir)
    app.doctreedir = to_unicode(app.doctreedir)

    #
    # The original make_id() targets HTML 4.01 and CSS1,
    # we need replace the function to use unicode directly in identifier.
    # (when the function returns empty string, docutils.nodes.document.set_id()
    #  makes 'id1' or so.)
    #
    docutils.nodes.make_id = my_make_id

    #
    # The function forces unicode to str with '%r', so we need replace.
    #
    sphinx.directives.other.TocTree.run = my_toctree_run

    return
    
