#
# -*- coding: utf-8 -*-

"""
    user_order_index_patch
    ~~~~~~~~~~~~~~~~~~~~~~

    The Sphinx extension to allow assign the order to sort terms in index.

    :copyright: (c) 2011, 2013, 2015 by Suzumizaki-Kimitaka(about Additional works)
    :license: 2-clause BSD.
"""

"""User ordered index hack for Sphinx.

This is the extension for for Sphinx( http://sphinx-doc.org/ ).
To use this, read the usage of extensions with/for Sphinx.
And make sure 'sort_order.py' and 'sort_order_XX.py' you want to use
can be imported from this extension.

You may also want to use the extension 'yogosyu.py' to re-order the items
in your 'genindex.html'. That extension gives you sort-key functionality
like MediaWiki, Wikipedia.

This extension hacks 'sphinx.environment.BuildEnvironment.create_index'.
No roles, directives, domains or etc. will be added.

Extension work is done by Suzumizaki-Kimitaka, May 2011.
Updated for Sphinx 1.2b3 is done December 2013.
Updated for Python 3.4.3 and Sphinx 1.3 at Mar 2015, 
"""

import bisect
import re
import string
import sys
import sphinx.environment
from six import text_type, iteritems
from itertools import groupby
from sphinx.util import split_into

# ========================================================
# Hacking helpers
# ========================================================

# Store 'SortOrderBase' or its subclass defined 'sort_order.py'.
_sort_order_obj = None

# The Dictionary from the entry-name to sort-key(yomi-gana)
_name_to_yomi = None

def yomi_or_given(s):
    """Return sort-key or given string if not found one
    
    @param s the string refer to one of entry
    @return the sort-key refer to the entry or just s if sort-key not found
    """
    s = s.lower()
    if s in _name_to_yomi:
        return _name_to_yomi[s]
    return s

def keyfunc(entry, lcletters=string.ascii_lowercase + '_'):
    """Sort the index entries

    @param entry the element to sort. entry[0] is the name of it.
    @param lcletters somewhat 'unrefactored' parameter.
    @return the comparable string to sort.

    Called by 'my_env_buildenv_create_index()' function.

    entry[0] is the name of the entry
    entry[1] is a list.
    entry[1][0] is a list of the urls include filename,
                extension and anchor to the entry.
    entry[1][1] is the dict of subwords.

    each entry is created in add_entry() in MyCreateIndex().
    
    Important: You should complete grouping entries by THIS function.
               Otherwise, keyfunc2 makes multi headings that have same name.
               (Generally, that work is done by calling
                '_sort_order_obj.get_string_to_sort()')
    """
    global _string_of_sort_order_to_yomi
    y = yomi_or_given(entry[0])
    r = _sort_order_obj.get_string_to_sort(y)
    return r

def keyfunc2(item, letters=string.ascii_uppercase + '_'):
    """Group the entries by letter.

    @param item key:value = the string to sort order: content
    @param letters somewhat 'unrefactored' parameter.
    @return group name in which (k, v) entry belongs.

    Called by 'my_env_buildenv_create_index()' function.
    
    See Important section written in the document of 'keyfunc()'.
    """
    k, v = item
    # hack: mutating the subitems dicts to a list in the keyfunc
    v[1] = sorted((si, se) for (si, (se, void)) in iteritems(v[1]))
    return _sort_order_obj.get_group_name(yomi_or_given(k))

# ========================================================
# Hacking function
# ========================================================

if sys.version_info.major < 3:
	_strs = basestring
else:
	_strs = str

def my_env_buildenv_create_index(self, builder, group_entries=True,
                                 _fixre=re.compile(r'(.*) ([(][^()]*[)])')):
        """Hack function to alter sphinx.environment.BuildEnvironment.create_index

        This hacking function is almost same except:
        - keyfunc and keyfunc2 are not inner function
        - setup _name_to_yomi dictionary at first
        """
        global _name_to_yomi
        _name_to_yomi = {}
        for k, v in self.domaindata['std']['objects'].items():
            if not isinstance(v[1], _strs):
                _name_to_yomi[k[1]] = v[1][1]
        new = {}

        def add_entry(word, subword, link=True, dic=new):
            # Force the word to be unicode if it's a ASCII bytestring.
            # This will solve problems with unicode normalization later.
            # For instance the RFC role will add bytestrings at the moment
            word = text_type(word)
            entry = dic.get(word)
            if not entry:
                dic[word] = entry = [[], {}]
            if subword:
                add_entry(subword, '', link=link, dic=entry[1])
            elif link:
                try:
                    uri = builder.get_relative_uri('genindex', fn) + '#' + tid
                except NoUri:
                    pass
                else:
                    # maintain links in sorted/deterministic order
                    bisect.insort(entry[0], (main, uri))

        for fn, entries in self.indexentries.items():
            # new entry types must be listed in directives/other.py!
            for type, value, tid, main in entries:
                try:
                    if type == 'single':
                        try:
                            entry, subentry = split_into(2, 'single', value)
                        except ValueError:
                            entry, = split_into(1, 'single', value)
                            subentry = ''
                        add_entry(entry, subentry)
                    elif type == 'pair':
                        first, second = split_into(2, 'pair', value)
                        add_entry(first, second)
                        add_entry(second, first)
                    elif type == 'triple':
                        first, second, third = split_into(3, 'triple', value)
                        add_entry(first, second+' '+third)
                        add_entry(second, third+', '+first)
                        add_entry(third, first+' '+second)
                    elif type == 'see':
                        first, second = split_into(2, 'see', value)
                        add_entry(first, _('see %s') % second, link=False)
                    elif type == 'seealso':
                        first, second = split_into(2, 'see', value)
                        add_entry(first, _('see also %s') % second, link=False)
                    else:
                        self.warn(fn, 'unknown index entry type %r' % type)
                except ValueError as err:
                    self.warn(fn, str(err))

        newlist = sorted(new.items(), key=keyfunc)

        if group_entries:
            # fixup entries: transform
            #   func() (in module foo)
            #   func() (in module bar)
            # into
            #   func()
            #     (in module foo)
            #     (in module bar)
            oldkey = ''
            oldsubitems = None
            i = 0
            while i < len(newlist):
                key, (targets, subitems) = newlist[i]
                # cannot move if it has subitems; structure gets too complex
                if not subitems:
                    m = _fixre.match(key)
                    if m:
                        if oldkey == m.group(1):
                            # prefixes match: add entry as subitem of the
                            # previous entry
                            oldsubitems.setdefault(m.group(2), [[], {}])[0].\
                                extend(targets)
                            del newlist[i]
                            continue
                        oldkey = m.group(1)
                    else:
                        oldkey = key
                oldsubitems = subitems
                i += 1

        return [(key_, list(group))
                for (key_, group) in groupby(newlist, keyfunc2)]

# ========================================================
# Initializing functions
# ========================================================

def determine_sort_order(app):
    """Determine sort order used in this module

    Called by Sphinx, on the event 'bulider-inited'.
    """
    global _sort_order_obj
    import sort_order
    _sort_order_obj = sort_order.get_sort_order(app.config)
    if _sort_order_obj.__class__ == sort_order.SortOrderLegacy:
        print ('user_ordered_index_patch.py: Using SortOrderLegacy.')
    return

def setup(app):
    """Extends Sphinx as we want

    @param app sphinx.application.Sphinx object to use add builder or so.
    """
    sphinx.environment.BuildEnvironment.create_index = my_env_buildenv_create_index
    app.connect('builder-inited', determine_sort_order)
    return
    
